const SoundDurations = {
	MenuTheme: {
		start: 120,
		duration: 21850
	},
	MasterPangoladin: {
		start: 0,
		duration: 41576,
		volume: .8
	},
	GameOverTheme: {
		start: 0,
		duration: 5325
	}
}

function SFX( fx ) {
	createjs.Sound.play( fx );
}

function Game( lib, root ) {
	
	this.lib = lib;
	this.stage = stage;
	this.stage.scaleX = this.stage.scaleY = 1/2;
	this.stage.canvas.width *= 1/2;
	this.stage.canvas.height *= 1/2;

	this.init();

	Object.defineProperties(this, {
		music: {
			get: function(){ return this._music },
			set: function( sound ) {
				if( this._music ) {
					this._music.stop()
					this._music.destroy();
				}
				this._music = createjs.Sound.play( sound, {
					startTime: SoundDurations[sound].start,
					duration: SoundDurations[sound].duration,
					volume: SoundDurations[sound].volume || 1,
					loop: -1
				});
			}
		}
	})
	
}
Game.prototype = {};
Game.prototype.init = function(){

	extendMenu( this.lib.Menu, this.lib );
	extendLevel( this.lib.Level, this.lib );
	extendHero( this.lib.Hero, this.lib );
	extendBall( this.lib.Ball, this.lib );
	extendEnemy( this.lib.Enemy, this.lib );
	extendUi( this.lib.Ui, this.lib );

	createjs.MovieClip.prototype.getBounds = function( optTarget ){

		var relativePos = new createjs.Point(0,0);
		if(optTarget){
			this.localToGlobal(0,0, relativePos);
			optTarget.globalToLocal( relativePos.x, relativePos.y, relativePos );
		}

		return {
			x : relativePos.x + this.nominalBounds.x*this.scaleX,
			y : relativePos.y + this.nominalBounds.y*this.scaleY,
			width : this.nominalBounds.width*this.scaleX,
			height : this.nominalBounds.height*this.scaleY
		};
	}
	createjs.MovieClip.prototype._reset = function(){
		this._rawPosition = -1;
		this._t = this.currentFrame = 0;
		// fuck off why you need to reset me motherfucker !!!!
		// this.paused = false;
	}

	this.ui = new this.lib.Ui();

}
Game.prototype.start = function(){

	var menu = new this.lib.Menu();
	this.music = 'MenuTheme';
	this.stage.addChild( menu );
	menu.addEventListener('start', function(){

		this.stage.removeChild(menu);
		menu = undefined;

		this.controller = new KeyboardController();

		this.music = 'MasterPangoladin';
		this.level = new this.lib.Level();
		this.level.bindController( this.controller );
		this.level.ui = this.ui;
		this.stage.addChild( this.level );
		this.stage.addChild( this.ui );
		this.level.addEventListener('end', this.end.bind(this));
		this.level.start();

	}.bind(this));

}
Game.prototype.end = function(){

	this.music = "GameOverTheme";
	this.stage.removeChild(this.level);
	var go = new this.lib.GameOver();
	go.score.text = this.level.score;
	this.stage.addChild(go);
	var onkeypress = function (e){
		if(e.key == 'Enter'){
			document.removeEventListener('keydown', onkeypress);
			this.stage.removeChild(go);
			this.start();
		}
	}.bind(this)
	document.addEventListener('keydown', onkeypress);
	//go.addEventListener( 'click', function(){
	//	this.stage.removeChild(go);
	//	this.start();
	//}.bind(this));

}

// Level
function extendLevel( Level, lib ){

	lib.Level = function(){
		Level.apply(this);
		Object.defineProperties(this, {
			left: {
				get: function(){ return 30 }
			},
			right: {
				get: function(){ return this.nominalBounds.width-30 }
			},
			top: {
				get: function(){ return 350 }
			},
			bottom: {
				get: function(){ return this.nominalBounds.height }
			}
		})
		this.balls = [];
		this.enemies = [];
	}
	lib.Level.prototype = Level.prototype;

	Level.prototype.start = function(){

		this.life = 2;
		this.score = 0;
		
		this.spawnEnemies(20);

		this.handleEvent = function(e){
			if(e.type == 'tick')
				this.update(e);
		}

		createjs.Ticker.addEventListener('tick', this );
	}

	Level.prototype.spawnEnemies = function(offsetY){
		const MAX_ENEMY = 5*14;
		offsetY = offsetY || 0;
		var numEnemy = 14;
		var enemyStart = this.left + 50;
		var enemySpace = (this.right - 50 - enemyStart)/(numEnemy-1);
		for( var i = 0; i<numEnemy && this.enemies.length < MAX_ENEMY; ++i ){
			var enemyType = [
				'type1',
				'type2'
			].reduce(function(acc, x){
				if(Math.random() < .5) return x;
				return acc;
			}, 'type1')
			var e = new lib.Enemy(enemyType);
			e.x = enemyStart + enemySpace*i;
			e.y = this.top - 20 + offsetY;
			e.alpha = 0;
			this.addChild(e);
			this.enemies.push( e );
		}
		this.lastSpawnEnemies = createjs.Ticker.getTime();
	}

	Level.prototype.update = function(e){

		var ds = e.delta/1000;

		// Hero Move
		this.hero.update(ds);
		var dir = 0;
		if(this.controller.left) dir = -1;
		if(this.controller.right) dir += 1;

		if(dir) {
			this.hero.speed += dir * this.hero.ACC * ds;
		}
		else if( this.hero.speed > 0 ) {
			this.hero.speed -= this.hero.FRICTION * ds;
			if(this.hero.speed < 0) this.hero.speed = 0;
		} else if( this.hero.speed < 0 ) {
			this.hero.speed += this.hero.FRICTION * ds;
			if(this.hero.speed > 0) this.hero.speed = 0;
		}
		speedDir = this.hero.speed < 0 ? -1 : (this.hero.speed > 0 ? 1 : 0);
		this.hero.speed = speedDir * 
			Math.min( this.hero.MAX_SPEED, Math.abs(this.hero.speed) );
		this.hero.x += this.hero.speed*ds;
		if(this.hero.x < this.left + 50) this.hero.x = this.left + 50;
		if(this.hero.x > this.right - 50) this.hero.x = this.right - 50;

		this.hero.sprite.paused = dir == 0;
		if( dir < 0 ) this.hero.sprite.scaleX = 1;
		else if( dir > 0 ) this.hero.sprite.scaleX = -1;

		// Hero Spell
		var spellKeys = {
			'a': 'damage'
		}
		for( var k in spellKeys ){
			if( this.controller.readPress(k) )
				this.hero.switchSpell( spellKeys[k] );
		}

		// Ball
		this.balls.forEach(function(ball){
			// Ball Move
			ball.x += ball.speedX*ds;
			ball.y += ball.speedY*ds;
			ball.rotation = 180 - 180*Math.atan2(ball.speedX, ball.speedY)/Math.PI;
			// Ball Collide Wall
			if(ball.x-ball.radius < this.left || ball.x+ball.radius > this.right){
				ball.x = Math.max(this.left+ball.radius, Math.min(ball.x, this.right-ball.radius));
				ball.speedX = -ball.speedX;
			}
			if(ball.y-ball.radius < this.top){ // margin to quick solve the touching enemies arriving
				ball.y = this.top + ball.radius;
				ball.speedY = -ball.speedY;
			} else if( ball.y > this.bottom ){
				this.balls.splice(this.balls.indexOf(ball), 1);
				this.removeChild(ball);
				--this.life;
				SFX('BallLose');
				return;
			}
			// Ball Collide Hero
			var ballBounds = ball.getCircleBounds(this);
			var collisionDirection = null;
			var heroBounds = this.hero.shield.getBounds(this);
			if( collisionDirection = rectCircleColliding( heroBounds, ballBounds ) )
			{
				ball.convert(this.hero.spell);
				this.hero.switchSpell('repel');
				switch(collisionDirection){
					case 'left':
						SFX('BallReboundHero');
						ball.x = heroBounds.x - ball.radius;
						var tmp = ball.speedX;
						ball.speedX = -Math.abs(ball.speedY);
						ball.speedY = -Math.abs(tmp);
						break;
					case 'right':
						SFX('BallReboundHero');
						ball.x = heroBounds.x + heroBounds.width + ball.radius;
						var tmp = ball.speedX;
						ball.speedX = Math.abs(ball.speedY);
						ball.speedY = -Math.abs(tmp);
						break;
					case 'top':
						SFX('BallReboundHero');
						ball.y = heroBounds.y - ball.radius;
						ball.speedY = -Math.abs(ball.speedY);
						ball.speedX += ball.x - this.hero.x;
						break;
					case 'bottom':
						break;
				}
			}
			// Ball Collide Enemies
			this.enemies.forEach(function(e, i){
				var collisionDirection = null;
				var enemyBounds = e.getBounds(this);
				if( collisionDirection = rectCircleColliding( enemyBounds, ballBounds ) )
				{
					this.score += 10;
					if( ball.type == 'repel' && collisionDirection == 'top' ) {
						if( ball.x-e.x < 0 ) collisionDirection = 'left';
						else collisionDirection = 'right';
					}
					SFX('EnemyHit');
					this.score += ball.strikeEnemy(e, collisionDirection);
					switch(collisionDirection){
						case 'left':
							ball.x = enemyBounds.x - ball.radius;
							ball.speedX = -Math.abs(ball.speedX);
							break;
						case 'right':
							ball.x = enemyBounds.x + enemyBounds.width + ball.radius;
							ball.speedX = Math.abs(ball.speedX);
							break;
						case 'top':
							ball.y = enemyBounds.y - ball.radius;
							ball.speedY = -Math.abs(ball.speedY);
							break;
						case 'bottom':
							ball.y = enemyBounds.y + enemyBounds.height + ball.radius;
							ball.speedY = Math.abs(ball.speedY);
							break;
					}
				}
			}.bind(this));
		}.bind(this));
		// New Ball
		if( this.balls.length == 0 ) {
			var newBall = this.hero.castBall();
			if(newBall) {
				this.balls.push( newBall );
				this.addChild( newBall );
			}
		}

		// Enemy
		var tintFor = function(e){
			return Math.min(Math.max(0, e.y - this.top), 100)/100;
		}.bind(this);
		this.enemies.forEach(function(e){
			e.update(ds);

			if( e.life == 0 ){
				this.enemies.splice(this.enemies.indexOf(e), 1);
				SFX('BallRebound');
				e.die(function(){
					this.removeChild(e);
				}.bind(this));
			}

			if( e.state == 'normal' ) {
				var enemyStop = this.top + (this.bottom-this.top) / 3;
				if( e.y < enemyStop ){
					var distToStop = 50 - Math.max(e.y-(enemyStop-50), 0);
					var speed = ds*Math.max( 1,
						e.speed*distToStop/50
					);
					e.y += speed;
				}
			} else if( e.state == 'repeled' ) {
				e.y -= 300*ds;
				e.rotation += 3600*ds;
				if( e.y < this.top ) {
					e.y = this.top;
					e.state = 'normal';
					e.rotation = 0;
				}
			}
			this.enemies.forEach(function(e2){
				if( e != e2 &&
					e.state != 'repeled' && e2.state != 'repeled'
				) {
					var eBounds = e.getBounds(this);
					var e2Bounds = e2.getBounds(this);
					if(	Math.abs(e2.x - e.x) < e2Bounds.width/2 &&
						e2.y - e.y >= 0 && // e is above e2
						e2Bounds.y - eBounds.y < e2Bounds.height
					)
						e.y = e2.y - e2Bounds.height;
				}
			}.bind(this));

			var tint = tintFor(e);
			if( e.tintValue != tint ) {
				e.tintValue = tint;
				var add = (1-tint)*0x29;
				e.filters = [
					new createjs.ColorFilter( tint, tint, tint, 1,
						add, add, add )
				];
				e.cache(e.nominalBounds.x, e.nominalBounds.y, e.nominalBounds.width, e.nominalBounds.height);
			}
			e.alpha = 1-(Math.max(Math.min(10, 10 - (e.y - (this.top -20))), 0)/10);

		}.bind(this));
		// spawn enemies
		function enemySpawnInterval() {
			return 20000;
		}
		if(createjs.Ticker.getTime() - this.lastSpawnEnemies > enemySpawnInterval())
			this.spawnEnemies();

		// display sort
		function z(c){
			if (c.name == 'background') return -100;
			if (c instanceof lib.Enemy && c.state == 'repeled') {
				return 100;
			}
			return 0;
		}
		this.children = this.children.sort(function(c1, c2){
			return (z(c1) - z(c2)) || (c1.y - c2.y);
		});

		// Ui
		this.ui.mana.fill.x = this.ui.mana.fill.nominalBounds.x - (1-this.hero.mana)*this.ui.mana.fill.nominalBounds.width;
		this.ui.score.counter.text = this.score;
		[this.ui.balls.ball1, this.ui.balls.ball2, this.ui.balls.ball3].forEach(
			function(ball, i){
				ball.visible = this.life > i;
			}.bind(this)
		);

		// end
		if(this.life < 0){
			this.end();
		}

	}

	Level.prototype.end = function() {
		createjs.Ticker.removeEventListener('tick', this );
		this.dispatchEvent( 'end' );
	}

	Level.prototype.bindController = function( controller ) {
		this.controller = controller;
	}

}

function extendUi( Ui, lib ) {

	lib.Ui = function(){
		Ui.apply(this);


	}
	lib.Ui.prototype = Ui.prototype;

}

function rectCircleColliding(rect, circle){

    var relX = circle.x - (rect.x+rect.width/2);
    var relY = circle.y - (rect.y+rect.height/2);
    var distX = Math.abs(relX);
    var distY = Math.abs(relY);

    if (distX > (rect.width/2 + circle.r)) { return 0; }
    if (distY > (rect.height/2 + circle.r)) { return 0; }

    if (distX <= (rect.width/2)) { return relY < 0 ? 'top' : 'bottom'; }
    if (distY <= (rect.height/2)) { return relX < 0 ? 'left' : 'right'; }

    var dx=distX-rect.width/2;
    var dy=distY-rect.height/2;

    if(dx*dx+dy*dy<=(circle.r*circle.r)){
    	if( dx/dy > rect.width/rect.height ) // closer in x, we have hit vertically
    		return relX < 0 ? 'left' : 'right';
    	else
    		return relY < 0 ? 'top' : 'bottom';
    }
    return 0;

}